﻿#include "stdafx.h"
#include "MQBasePlugin.h"
#include "MQ3DLib.h"
#include <string>
#include <vector>
#include <map>

const char unattached_material_name[] = "_Unattached";


static std::string AnsiToUtf8(const char *ptr)
{
	if(ptr == NULL) return std::string();

	int lenw = MultiByteToWideChar(CP_ACP,0,ptr,-1,NULL,0);
	wchar_t *strw = (wchar_t*)malloc(sizeof(wchar_t) * (lenw+1));
	MultiByteToWideChar(CP_ACP,0,ptr,-1,strw,lenw);

	int lenu = WideCharToMultiByte(CP_UTF8,0,strw,-1,NULL,0,NULL,NULL);
	char *stru = (char*)malloc(lenu+1);
	WideCharToMultiByte(CP_UTF8,0,strw,-1,stru,lenu,NULL,NULL);

	free(strw);

	std::string str(stru);
	free(stru);
	return str;
}

std::wstring AnsiToWide(const char *ptr)
{
	if(ptr == NULL) return std::wstring();

	int lenw = MultiByteToWideChar(CP_ACP,0,ptr,-1,NULL,0);
	wchar_t *strw = (wchar_t*)malloc(sizeof(wchar_t) * (lenw+1));
	MultiByteToWideChar(CP_ACP,0,ptr,-1,strw,lenw);

	std::wstring str(strw);
	free(strw);
	return str;
}

std::string WideToUtf8(const wchar_t *ptr)
{
	if(ptr == NULL) return std::string();

	int lenu = WideCharToMultiByte(CP_UTF8,0,ptr,-1,NULL,0,NULL,NULL);
	char *stru = (char*)malloc(lenu+1);
	WideCharToMultiByte(CP_UTF8,0,ptr,-1,stru,lenu,NULL,NULL);

	std::string str(stru);
	free(stru);
	return str;
}


class ColladaExporter : public MQExportPlugin
{
public:
	ColladaExporter();
	~ColladaExporter();

	// プラグインIDを返す。
	virtual void GetPlugInID(DWORD *Product, DWORD *ID);

	// プラグイン名を返す。
	virtual const char *GetPlugInName(void);

	// 入力または出力可能なファイルタイプを返す。
	virtual const char *EnumFileType(int index);

	// 入力または出力可能な拡張子を返す。
	virtual const char *EnumFileExt(int index);

	// ファイルの出力
	virtual BOOL ExportFile(int index, const char *filename, MQDocument doc);

private:
	BOOL ExportDAE(const char *filename, MQDocument doc);
	void ExportObjectNode(FILE *fh, MQDocument doc, MQObject obj, const std::vector<std::string>& mat_name_utf8s);

	bool m_bLocal;
};


ColladaExporter::ColladaExporter()
{
	m_bLocal = true;
}

ColladaExporter::~ColladaExporter()
{
}

// プラグインIDを返す。
void ColladaExporter::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0x56A31D20;
	*ID = 0x5241E4A3;
}

// プラグイン名を返す。
const char *ColladaExporter::GetPlugInName(void)
{
	return "RIB Exporter   Copyright(C) 2011, O.Mizno";
}


// 入力または出力可能なファイルタイプを返す。
const char *ColladaExporter::EnumFileType(int index)
{
	switch(index){
	case 0: return "COLLADA (*.dae)";
	}
	return NULL;
}

// 入力または出力可能な拡張子を返す。
const char *ColladaExporter::EnumFileExt(int index)
{
	switch(index){
	case 0: return "dae";
	}
	return NULL;
}

// ファイルの出力
BOOL ColladaExporter::ExportFile(int index, const char *filename, MQDocument doc)
{
	if(index == 0){
		return ExportDAE(filename, doc);
	}
	return FALSE;
}

BOOL ColladaExporter::ExportDAE(const char *filename, MQDocument doc)
{
	FILE *fh = NULL;
	errno_t val = fopen_s(&fh, filename, "w");
	if(val != 0){
		return FALSE;
	}

	SYSTEMTIME tm;
	GetSystemTime(&tm);

	fprintf(fh, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(fh, "<COLLADA xmlns=\"http://www.collada.org/2005/11/COLLADASchema\" version=\"1.4.1\">\n");
	//fprintf(fh, "<COLLADA xmlns=\"http://www.collada.org/2008/03/COLLADASchema\" version=\"1.5.0\">\n");
	fprintf(fh, " <asset>\n");
	fprintf(fh, "  <contributor>\n");
	fprintf(fh, "   <authoring_tool>ExpCollada for Metasequoia</authoring_tool>\n");
	fprintf(fh, "  </contributor>\n");
	fprintf(fh, "  <created>%04d-%02d-%02dT%02d:%02d:%02dZ</created>\n", tm.wYear, tm.wMonth, tm.wDay, tm.wHour, tm.wMinute, tm.wSecond);
	fprintf(fh, "  <modified>%04d-%02d-%02dT%02d:%02d:%02dZ</modified>\n", tm.wYear, tm.wMonth, tm.wDay, tm.wHour, tm.wMinute, tm.wSecond);
	fprintf(fh, "  <up_axis>Y_UP</up_axis>\n");
	fprintf(fh, " </asset>\n");

	MQScene scene = doc->GetScene(0);
	
	fprintf(fh, " <library_cameras>\n");
	fprintf(fh, "  <camera id=\"Camera-Camera\" name=\"Camera-Camera\">\n");
	fprintf(fh, "   <optics>\n");
	fprintf(fh, "    <technique_common>\n");
	fprintf(fh, "     <perspective>\n");
	fprintf(fh, "      <yfov>%.3f</yfov>\n", scene->GetFOV() / PI * 180);
	fprintf(fh, "      <znear>0.1</znear>\n");
	fprintf(fh, "      <zfar>10000.0</zfar>\n");
	fprintf(fh, "     </perspective>\n");
	fprintf(fh, "    </technique_common>\n");
	fprintf(fh, "   </optics>\n");
	fprintf(fh, "  </camera>\n");
	fprintf(fh, " </library_cameras>\n");

	int numObj = doc->GetObjectCount();
	int numMat = doc->GetMaterialCount();
	
	// Check if the material is used
	std::vector<bool> material_used(numMat+1, false);
	for(int obj_idx = 0; obj_idx < numObj; obj_idx++)
	{
		MQObject obj = doc->GetObject(obj_idx);
		if(obj == NULL){
			continue;
		}

		int numFace = obj->GetFaceCount();
		for(int face_idx = 0; face_idx < numFace; face_idx++)
		{
			int pt_num = obj->GetFacePointCount(face_idx);
			if(pt_num > 0){
				int mat_idx = obj->GetFaceMaterial(face_idx);
				if(mat_idx >= 0 && mat_idx < numMat){
					material_used[mat_idx] = true;
				}else{
					material_used[numMat] = true;
				}
			}
		}
	}

	// Count used materials
	std::vector<std::string> mat_name_utf8s(numMat+1);
	int mat_used_num = 0;
	for(int mat_idx = 0; mat_idx <= numMat; mat_idx++){
		if(material_used[mat_idx]){
			mat_used_num++;

			// Get a material name
			char mat_name[64];
			if(mat_idx < numMat){
				MQMaterial mat = doc->GetMaterial(mat_idx);
				if(mat == NULL) continue;

				mat->GetName(mat_name, _countof(mat_name));
			}else{
				strcpy_s(mat_name, sizeof(mat_name), unattached_material_name);
			}

			mat_name_utf8s[mat_idx] = AnsiToUtf8(mat_name);
		}
	}

	// Images
	std::map<std::string, std::string> texture_map; // <filename(ANSI), id(UTF-8)>
	int tex_count = 0;
	for(int mat_idx = 0; mat_idx < numMat; mat_idx++)
	{
		// Unused material is not outputted
		if(!material_used[mat_idx]) continue;

		MQMaterial mat = doc->GetMaterial(mat_idx);
		if(mat == NULL) continue;

		char map_name[MAX_PATH];
		mat->GetTextureName(map_name, MAX_PATH);
		if(map_name[0] != '\0'){
			std::wstring map_wstr = AnsiToWide(map_name);
			for(size_t i=0; i<map_wstr.length(); i++){
				if(map_wstr[i] == L'\\') map_wstr[i] = L'/';
			}
			std::string map_name_utf8 = WideToUtf8(map_wstr.c_str());

			if(texture_map.find(map_name) == texture_map.end()){
				char tex_id[20];
				sprintf_s(tex_id, sizeof(tex_id), "texture%d", ++tex_count);
				std::string tex_id_utf8 = AnsiToUtf8(tex_id);

				texture_map.insert(std::make_pair<std::string, std::string>(map_name, tex_id_utf8));
			
				if(tex_count == 1){
					fprintf(fh, " <library_images>\n");
				}
				fprintf(fh, "  <image id=\"%s\" name=\"%s\">\n", tex_id_utf8.c_str(), tex_id_utf8.c_str());
				fprintf(fh, "   <init_from>%s</init_from>\n", map_name_utf8.c_str());
				fprintf(fh, "  </image>\n");
			}
		}
	}
	if(tex_count > 0){
		fprintf(fh, " </library_images>\n");
	}

	// Effects
	if(mat_used_num > 0){
		fprintf(fh, " <library_effects>\n");
	}
	for(int mat_idx = 0; mat_idx <= numMat; mat_idx++)
	{
		// Unused material is not outputted
		if(!material_used[mat_idx]) continue;

		MQMaterial mat;
		if(mat_idx < numMat){
			mat = doc->GetMaterial(mat_idx);
			if(mat == NULL) continue;
		}else{
			// Create a temporary material for an unattached material
			mat = MQ_CreateMaterial();
		}

		std::string mat_name_utf8 = mat_name_utf8s[mat_idx];

		MQColor col = mat->GetColor();

		fprintf(fh, "  <effect id=\"%s-fx\" name=\"%s-fx\">\n", mat_name_utf8.c_str(), mat_name_utf8.c_str());
		fprintf(fh, "   <profile_COMMON>\n");

		char map_name[MAX_PATH];
		mat->GetTextureName(map_name, MAX_PATH);
		if(map_name[0] != '\0'){
			std::map<std::string, std::string>::iterator it = texture_map.find(map_name);
			if(it != texture_map.end()){
				fprintf(fh, "    <newparam sid=\"%s-surface\">\n", (*it).second.c_str());
				fprintf(fh, "     <surface type=\"2D\">\n");
				fprintf(fh, "      <init_from>%s</init_from>\n", (*it).second.c_str());
				//fprintf(fh, "      <format>R8G8B8A8</format>\n");
				fprintf(fh, "     </surface>\n");
				fprintf(fh, "    </newparam>\n");
				fprintf(fh, "    <newparam sid=\"%s-sampler\">\n", (*it).second.c_str());
				fprintf(fh, "    <sampler2D>\n");
				fprintf(fh, "     <source>%s-surface</source>\n", (*it).second.c_str());
				fprintf(fh, "     <minfilter>LINEAR</minfilter>\n");
				fprintf(fh, "     <magfilter>LINEAR</magfilter>\n");
				fprintf(fh, "    </sampler2D>\n");
				fprintf(fh, "   </newparam>\n");
			}
		}

		fprintf(fh, "    <technique sid=\"metaseq\">\n");

		int shader = mat->GetShader();
		bool valid_amb = true, valid_dif = true, valid_spc = true;
		switch(shader){
		case MQMATERIAL_SHADER_CONSTANT:
			fprintf(fh, "     <constant>\n");
			valid_amb = false;
			valid_dif = false;
			valid_spc = false;
			break;
		case MQMATERIAL_SHADER_LAMBERT:
			fprintf(fh, "     <lambert>\n");
			valid_spc = false;
			break;
		case MQMATERIAL_SHADER_BLINN:
			fprintf(fh, "     <blinn>\n");
			break;
		case MQMATERIAL_SHADER_PHONG:
		case MQMATERIAL_SHADER_CLASSIC:
		default:
			fprintf(fh, "     <phong>\n");
			break;
		}
		
		float emi = mat->GetEmission();
		fprintf(fh, "      <emission>\n");
		fprintf(fh, "       <color>%.3f %.3f %.3f 1</color>\n", col.r*emi, col.g*emi, col.b*emi);
		fprintf(fh, "      </emission>\n");

		if(valid_amb){
			float amb = mat->GetAmbient();
			fprintf(fh, "      <ambient>\n");
			fprintf(fh, "       <color>%.3f %.3f %.3f 1</color>\n", amb, amb, amb);
			fprintf(fh, "      </ambient>\n");
		}

		if(valid_dif){
			float dif = mat->GetDiffuse();
			fprintf(fh, "      <diffuse>\n");
			if(map_name[0] != '\0'){
				std::map<std::string, std::string>::iterator it = texture_map.find(map_name);
				if(it != texture_map.end()){
					fprintf(fh, "       <texture texture=\"%s-sampler\" texcoord=\"UVSET0\"/>\n", (*it).second.c_str());
				}
			}else{
				fprintf(fh, "       <color>%.3f %.3f %.3f 1</color>\n", col.r*dif, col.g*dif, col.b*dif);
			}
			fprintf(fh, "      </diffuse>\n");
		}
		
		if(valid_spc){
			float spc = mat->GetSpecular();
			fprintf(fh, "      <specular>\n");
			fprintf(fh, "       <color>%.3f %.3f %.3f 1</color>\n", spc, spc, spc);
			fprintf(fh, "      </specular>\n");
			
			float power = mat->GetPower();
			fprintf(fh, "      <shininess>\n");
			fprintf(fh, "       <float>%.3f</float>\n", power);
			fprintf(fh, "      </shininess>\n");
		}

		fprintf(fh, "      <reflective>\n");
		fprintf(fh, "       <color>1 1 1 1</color>\n");
		fprintf(fh, "      </reflective>\n");

		fprintf(fh, "      <reflectivity>\n");
		fprintf(fh, "       <float>0.0</float>\n");
		fprintf(fh, "      </reflectivity>\n");

		float alpha = mat->GetAlpha();
		fprintf(fh, "      <transparent>\n");
		fprintf(fh, "       <color>1 1 1 1</color>\n");
		fprintf(fh, "      </transparent>\n");
		fprintf(fh, "      <transparency>\n");
		fprintf(fh, "       <float>%.3f</float>\n", alpha);
		fprintf(fh, "      </transparency>\n");

		switch(shader){
		case MQMATERIAL_SHADER_CONSTANT:
			fprintf(fh, "     </constant>\n");
			break;
		case MQMATERIAL_SHADER_LAMBERT:
			fprintf(fh, "     </lambert>\n");
			break;
		case MQMATERIAL_SHADER_BLINN:
			fprintf(fh, "     </blinn>\n");
			break;
		case MQMATERIAL_SHADER_PHONG:
		case MQMATERIAL_SHADER_CLASSIC:
		default:
			fprintf(fh, "     </phong>\n");
			break;
		}
		fprintf(fh, "    </technique>\n");
		fprintf(fh, "   </profile_COMMON>\n");
		fprintf(fh, "  </effect>\n");

		// Delete a temporary material for an unattached material
		if(mat_idx == numMat){
			mat->DeleteThis();
		}
	}
	if(mat_used_num > 0){
		fprintf(fh, " </library_effects>\n");
	}

	// Materials
	if(mat_used_num > 0){
		fprintf(fh, " <library_materials>\n");
	}
	for(int mat_idx = 0; mat_idx <= numMat; mat_idx++)
	{
		// Unused material is not outputted
		if(!material_used[mat_idx]) continue;

		std::string mat_name_utf8 = mat_name_utf8s[mat_idx];

		fprintf(fh, "  <material id=\"%s\" name=\"%s\">\n", mat_name_utf8.c_str(), mat_name_utf8.c_str());
		fprintf(fh, "   <instance_effect url=\"#%s-fx\"/>\n", mat_name_utf8.c_str());
		fprintf(fh, "  </material>\n");
	}
	if(mat_used_num > 0){
		fprintf(fh, " </library_materials>\n");
	}

	// Geometries
	if(numObj > 0){
		fprintf(fh, " <library_geometries>\n");
	}
	for(int obj_idx = 0; obj_idx < numObj; obj_idx++)
	{
		MQObject obj = doc->GetObject(obj_idx);
		if(obj == NULL){
			continue;
		}

		char obj_name[64];
		obj->GetName(obj_name, _countof(obj_name));
		std::string obj_name_utf8 = AnsiToUtf8(obj_name);

		int numVert = obj->GetVertexCount();
		int numFace = obj->GetFaceCount();

		std::vector<int> apex_index;
		int numApex = 0;
		for(int face_idx = 0; face_idx < numFace; face_idx++)
		{
			int pt_num = obj->GetFacePointCount(face_idx);
			apex_index.push_back(numApex);
			numApex += pt_num;
		}

		MQMatrix invmat;
		doc->GetGlobalInverseMatrix(obj, invmat);

		MQObjNormal obj_normal(obj);

		int normal_current = -1;
		for(int face_idx = 0; face_idx < numFace; face_idx++)
		{
			int pt_num = obj->GetFacePointCount(face_idx);
			MQPoint prev_nrm;
			for(int pt = 0; pt < pt_num; pt++){
				MQPoint nrm = obj_normal.Get(face_idx, pt);
				if(pt == 0 || prev_nrm != nrm){
					prev_nrm = nrm;
					normal_current++;
				}
			}
		}

		fprintf(fh, "  <geometry id=\"%s-lib\" name=\"%s\">\n", obj_name_utf8.c_str(), obj_name_utf8.c_str());
		fprintf(fh, "   <mesh>\n");
		
		// Positions
		fprintf(fh, "    <source id=\"%s-lib-positions\" name=\"position\">\n", obj_name_utf8.c_str());
		fprintf(fh, "     <float_array id=\"%s-lib-positions-array\" count=\"%d\">", obj_name_utf8.c_str(), numVert * 3);
		for(int vert_idx = 0; vert_idx < numVert; vert_idx++)
		{
			MQPoint pos = obj->GetVertex(vert_idx);
			if(vert_idx > 0) fprintf(fh, " ");
			if(m_bLocal) pos = pos * invmat;
			fprintf(fh, "%.5f %.5f %.5f", pos.x, pos.y, pos.z);

			//if(vert_idx % 10 == 9) fprintf(fh, "\n     ");
		}
		fprintf(fh, "</float_array>\n");
		fprintf(fh, "     <technique_common>\n");
		fprintf(fh, "      <accessor count=\"%d\" source=\"#%s-lib-positions-array\" stride=\"3\">\n", numVert, obj_name_utf8.c_str());
		fprintf(fh, "       <param name=\"X\" type=\"float\"/>\n");
		fprintf(fh, "       <param name=\"Y\" type=\"float\"/>\n");
		fprintf(fh, "       <param name=\"Z\" type=\"float\"/>\n");
		fprintf(fh, "      </accessor>\n");
		fprintf(fh, "     </technique_common>\n");
		fprintf(fh, "    </source>\n");

		// Normals
		fprintf(fh, "    <source id=\"%s-lib-normals\" name=\"normal\">\n", obj_name_utf8.c_str());
		fprintf(fh, "     <float_array id=\"%s-lib-normals-array\" count=\"%d\">", obj_name_utf8.c_str(), (normal_current+1) * 3);
		std::vector<int> normal_indices;
		normal_current = -1;
		for(int face_idx = 0; face_idx < numFace; face_idx++)
		{
			int pt_num = obj->GetFacePointCount(face_idx);
			MQPoint prev_nrm;
			for(int pt = 0; pt < pt_num; pt++){
				MQPoint nrm = obj_normal.Get(face_idx, pt);
				if(m_bLocal){
					nrm = invmat.Mult3(nrm);
					nrm.normalize();
				}
				if(pt == 0 || prev_nrm != nrm){
					if(normal_current >= 0) fprintf(fh, " ");
					fprintf(fh, "%.5f %.5f %.5f", nrm.x, nrm.y, nrm.z);
					prev_nrm = nrm;
					normal_current++;
				}
				normal_indices.push_back(normal_current);
			}
		}
		fprintf(fh, "</float_array>\n");
		fprintf(fh, "     <technique_common>\n");
		fprintf(fh, "      <accessor count=\"%d\" source=\"#%s-lib-normals-array\" stride=\"3\">\n", normal_current+1, obj_name_utf8.c_str());
		fprintf(fh, "       <param name=\"X\" type=\"float\"/>\n");
		fprintf(fh, "       <param name=\"Y\" type=\"float\"/>\n");
		fprintf(fh, "       <param name=\"Z\" type=\"float\"/>\n");
		fprintf(fh, "      </accessor>\n");
		fprintf(fh, "     </technique_common>\n");
		fprintf(fh, "    </source>\n");
		
		// TexCoords
		fprintf(fh, "    <source id=\"%s-lib-texcoords\">\n", obj_name_utf8.c_str());
		fprintf(fh, "     <float_array id=\"%s-lib-texcoords-array\" count=\"%d\">", obj_name_utf8.c_str(), numApex * 2);
		for(int face_idx = 0; face_idx < numFace; face_idx++)
		{
			int pt_num = obj->GetFacePointCount(face_idx);
			MQCoordinate coord[4];
			obj->GetFaceCoordinateArray(face_idx, coord);
			for(int pt = 0; pt < pt_num; pt++){
				if(!(face_idx == 0 && pt == 0)) fprintf(fh, " ");
				fprintf(fh, "%.5f %.5f", coord[pt].u, coord[pt].v);
			}
		}
		fprintf(fh, "</float_array>\n");
		fprintf(fh, "     <technique_common>\n");
		fprintf(fh, "      <accessor count=\"%d\" offset=\"0\" source=\"#%s-lib-texcoords-array\" stride=\"2\">\n", numApex, obj_name_utf8.c_str());
		fprintf(fh, "       <param name=\"U\" type=\"float\"/>\n");
		fprintf(fh, "       <param name=\"V\" type=\"float\"/>\n");
		fprintf(fh, "      </accessor>\n");
		fprintf(fh, "     </technique_common>\n");
		fprintf(fh, "    </source>\n");
		
		// Vertices
		fprintf(fh, "    <vertices id=\"%s-lib-vertices\">\n", obj_name_utf8.c_str());
		fprintf(fh, "     <input semantic=\"POSITION\" source=\"#%s-lib-positions\"/>\n", obj_name_utf8.c_str());
		fprintf(fh, "    </vertices>\n");

		// Polylist
		//fprintf(fh, "    <polylist count=\"%d\" material=\"BlueSG\">\n", numFace);
		std::vector<bool> face_done(numFace, false);
		int start_face_idx = 0;
		int last_mat_idx = -2;
		for(int start_face_idx = 0; start_face_idx < numFace; start_face_idx++)
		{
			if(face_done[start_face_idx]) continue;

			int pt_num = obj->GetFacePointCount(start_face_idx);
			if(pt_num == 0) continue;

			// Count faces with same material
			std::vector<bool> target(numFace, false);
			int mat_idx = obj->GetFaceMaterial(start_face_idx);
			int face_count = 0;
			for(int face_idx = start_face_idx; face_idx < numFace; face_idx++)
			{
				if(face_done[face_idx]) continue;
					
				int pt_num = obj->GetFacePointCount(face_idx);
				if(pt_num == 0) continue;

				if(mat_idx == obj->GetFaceMaterial(face_idx)){
					face_count++;
					target[face_idx] = face_done[face_idx] = true;
				}
			}

			// Output polygon list
			int mat_array_idx = (mat_idx >= 0) ? mat_idx : numMat;
			std::string mat_name_utf8 = mat_name_utf8s[mat_array_idx];
			fprintf(fh, "    <polylist count=\"%d\" material=\"%s\">\n", face_count, mat_name_utf8.c_str());
			fprintf(fh, "     <input offset=\"0\" semantic=\"VERTEX\" source=\"#%s-lib-vertices\"/>\n", obj_name_utf8.c_str());
			fprintf(fh, "     <input offset=\"1\" semantic=\"NORMAL\" source=\"#%s-lib-normals\"/>\n", obj_name_utf8.c_str());
			fprintf(fh, "     <input offset=\"2\" semantic=\"TEXCOORD\" source=\"#%s-lib-texcoords\"/>\n", obj_name_utf8.c_str());
			fprintf(fh, "     <vcount>");
			for(int face_idx = start_face_idx; face_idx < numFace; face_idx++)
			{
				if(target[face_idx]){
					int pt_num = obj->GetFacePointCount(face_idx);
					if(face_idx > start_face_idx) fprintf(fh, " ");
					fprintf(fh, "%d", pt_num);
				}
			}
			fprintf(fh, "</vcount>\n");
			fprintf(fh, "     <p>");
			for(int face_idx = start_face_idx; face_idx < numFace; face_idx++)
			{
				if(target[face_idx]){
					int apex_idx = apex_index[face_idx];
					int pt_num = obj->GetFacePointCount(face_idx);
					int vertex[4];
					obj->GetFacePointArray(face_idx, vertex);
					for(int pt = 0; pt < pt_num; pt++){
						int idx = pt_num - 1 - pt;
						if(!(face_idx == start_face_idx && pt == 0)) fprintf(fh, " ");
						fprintf(fh, "%d", vertex[idx]); // VERTEX
						fprintf(fh, " %d", normal_indices[apex_idx + idx]); // NORMAL
						fprintf(fh, " %d", apex_idx + idx); // TEXCOORD
					}
				}
			}
			fprintf(fh, "</p>\n");
			fprintf(fh, "    </polylist>\n");
		}
		fprintf(fh, "   </mesh>\n");
		fprintf(fh, "  </geometry>\n");
	}
	if(numObj > 0){
		fprintf(fh, " </library_geometries>\n");
	}

	fprintf(fh, " <library_visual_scenes>\n");
	fprintf(fh, "  <visual_scene id=\"Scene\" name=\"Scene\">\n");
	
	MQPoint cam_pos = scene->GetCameraPosition();
	MQAngle cam_angle = scene->GetCameraAngle();
	fprintf(fh, "   <node id=\"Camera\" name=\"Camera\">\n");
	fprintf(fh, "    <rotate sid=\"rotateY\">0 1 0 %.5f</rotate>\n", -cam_angle.head);
	fprintf(fh, "    <rotate sid=\"rotateX\">1 0 0 %.5f</rotate>\n", -cam_angle.pich);
	fprintf(fh, "    <rotate sid=\"rotateZ\">0 0 1 %.5f</rotate>\n", -cam_angle.bank);
	fprintf(fh, "    <translate sid=\"translate\">%.5f %.5f %.5f</translate>\n", 0.0f, 0.0f, cam_pos.abs());
	//fprintf(fh, "    <translate sid=\"translate\">%.5f %.5f %.5f</translate>\n", cam_pos.x, cam_pos.y, cam_pos.z);
	//fprintf(fh, "    <scale sid=\"scale\">1.00000 1.00000 1.00000</scale>\n");
	fprintf(fh, "    <instance_camera url=\"#Camera-Camera\"/>\n");
	fprintf(fh, "   </node>\n");
	
	for(int obj_idx = 0; obj_idx < numObj; obj_idx++)
	{
		MQObject obj = doc->GetObject(obj_idx);
		if(obj == NULL){
			continue;
		}

		if(obj->GetDepth() == 0){
			ExportObjectNode(fh, doc, obj, mat_name_utf8s);
		}
	}
	
	fprintf(fh, "  </visual_scene>\n");
	fprintf(fh, " </library_visual_scenes>\n");

	fprintf(fh, " <scene>\n");
	fprintf(fh, "  <instance_visual_scene url=\"#Scene\"/>\n");
	fprintf(fh, " </scene>\n");
	fprintf(fh, "</COLLADA>\n");

	int ret = fclose(fh);
	return (ret == 0) ? TRUE : FALSE;
}


void ColladaExporter::ExportObjectNode(FILE *fh, MQDocument doc, MQObject obj, const std::vector<std::string>& mat_name_utf8s)
{
	if(obj == NULL){
		return;
	}

	int numMat = doc->GetMaterialCount();

	char obj_name[64];
	obj->GetName(obj_name, _countof(obj_name));
	std::string obj_name_utf8 = AnsiToUtf8(obj_name);

	int depth = obj->GetDepth();
	std::string head;
	for(int i = 0; i < depth + 3; i++){
		head += " ";
	}

	fprintf(fh, "%s<node id=\"%s\" name=\"%s\">\n", head.c_str(), obj_name_utf8.c_str(), obj_name_utf8.c_str());

	// Local coordinate
	if(m_bLocal){
		MQPoint scale = obj->GetScaling();
		MQAngle rotate = obj->GetRotation();
		MQPoint trans = obj->GetTranslation();
		fprintf(fh, "%s <translate sid=\"translate\">%.5f %.5f %.5f</translate>\n", head.c_str(), trans.x, trans.y, trans.z);
		fprintf(fh, "%s <rotate sid=\"head\">0 1 0 %.5f</rotate>\n", head.c_str(), rotate.head);
		fprintf(fh, "%s <rotate sid=\"pitch\">1 0 0 %.5f</rotate>\n", head.c_str(), rotate.pich);
		fprintf(fh, "%s <rotate sid=\"bank\">0 0 1 %.5f</rotate>\n", head.c_str(), rotate.bank);
		fprintf(fh, "%s <scale sid=\"scale\">%.5f %.5f %.5f</scale>\n", head.c_str(), scale.x, scale.y, scale.z);
	}

	// Geometry
	fprintf(fh, "%s <instance_geometry url=\"#%s-lib\">\n", head.c_str(), obj_name_utf8.c_str());
	fprintf(fh, "%s  <bind_material>\n", head.c_str());
	fprintf(fh, "%s   <technique_common>\n", head.c_str());
	
	int numFace = obj->GetFaceCount();
	std::vector<bool> mat_done(numMat+1, false);
	for(int face_idx = 0; face_idx < numFace; face_idx++)
	{
		int pt_num = obj->GetFacePointCount(face_idx);
		if(pt_num == 0) continue;

		int mat_idx = obj->GetFaceMaterial(face_idx);
		int mat_array_idx = (mat_idx >= 0) ? mat_idx : numMat;
		if(!mat_done[mat_array_idx]){
			std::string mat_name_utf8 = mat_name_utf8s[mat_array_idx];

			fprintf(fh, "%s    <instance_material symbol=\"%s\" target=\"#%s\">\n", head.c_str(), mat_name_utf8.c_str(), mat_name_utf8.c_str());
			fprintf(fh, "%s     <bind_vertex_input semantic=\"UVSET0\" input_semantic=\"TEXCOORD\"/>\n", head.c_str());
			fprintf(fh, "%s    </instance_material>\n", head.c_str());

			mat_done[mat_array_idx] = true;
		}
	}
	
	fprintf(fh, "%s   </technique_common>\n", head.c_str());
	fprintf(fh, "%s  </bind_material>\n", head.c_str());
	fprintf(fh, "%s </instance_geometry>\n", head.c_str());
	
	// Output children
	int numChild = doc->GetChildObjectCount(obj);
	for(int i = 0; i < numChild; i++){
		MQObject child = doc->GetChildObject(obj, i);
		ExportObjectNode(fh, doc, child, mat_name_utf8s);
	}

	// End a node
	fprintf(fh, "%s</node>\n", head.c_str());
}


MQBasePlugin *GetPluginClass()
{
	static ColladaExporter plugin;
	return &plugin;
}

