﻿//---------------------------------------------------------------------------
//
//   ImportMQO
//
//          Copyright(C) 2000-2012, O.Mizno
//
//     Sample of Import plug-in for importing a MQO (Metasequoia) file.
//     Created DLL must be installed in "Plugins\Import" directory.
//     See a document about the MQO file format on the official Web site.
//    Some procedures, such as TrialNoise, are omitted in this sample. And, 
//    functions not supported by plug-in API are commented out.
//     This is just a sample. Metasequoia already contains the embedded 
//    MQO importer in the main body. There is no means to use this plug-in.
//
//    　MQO(Metasequoia)形式を入力するインポートプラグインのサンプル。
//    　作成したDLLは"Plugins\Import"フォルダに入れる必要がある。
//      MQOファイルフォーマットについては別途ホームページ上の情報を参照
//    のこと。このコード自体は内部で使われているコードを少し手直ししたも
//    のであるが、TrialNoiseなどのいくつかのチャンクの処理は割愛している。
//    プラグインAPIがサポートしていない機能はコメントアウトされている。
//    　これは単なるサンプルであり、Metasequoiaの内部にはMQOインポート機
//    能が組み込まれているので、このプロジェクトで作成したプラグインを使
//    う意味はない。
//
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "MQPlugin.h"
#include "Strings.h"

#define COMCHECK(i,s) 	afs->chkcom(s,i)
#define CMPSTR(i,s) 	afs->chkstr(i,s)


#define LOADERMQO_SUCCEEDED       0
#define LOADERMQO_ERROR          -1
#define LOADERMQO_EOF            -2
#define LOADERMQO_HEADER_ERR     -3
#define LOADERMQO_FORMAT_ERR     -4
#define LOADERMQO_TRIAL_NOISE    -5


class LoaderMQO
{
protected:
	FILE *fh;
	AsciiFileString *afs;

	MQDocument doc;

	bool CheckHeader(void);
	bool IgnoreUnknown(int string_count);
	bool IgnoreBinary(int string_count);
	bool MainLoop(void);

	bool LoadScene(void);
	bool LoadMaterial(void);
	bool LoadObject(MQObject obj);
	bool LoadVertex(MQObject obj);
	bool LoadVertexAttr(MQObject obj);
	bool LoadVertexAttrColor(MQObject obj);
	bool LoadVertexAttrWeight(MQObject obj);
	bool LoadBVertex(MQObject obj);
	bool LoadFace(MQObject obj);

public:
	LoaderMQO(MQDocument document, FILE *handle);
	virtual ~LoaderMQO();

	int Load(void);
};


BOOL LoadMQO(const char *filename, MQDocument doc);


//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	//プラグインとしては特に必要な処理はないので、何もせずにTRUEを返す
    return TRUE;
}

//---------------------------------------------------------------------------
//  MQGetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT void MQGetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x543AEF01;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQGetPlugInName(void)
{
	// プラグイン名
	return "MQO Importer   Copyright(C) 2000-2012, O.Mizno";
}

//---------------------------------------------------------------------------
//  MQGetPlugInType
//    プラグインのタイプを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT int MQGetPlugInType(void)
{
	// 入力用プラグインである
	return MQPLUGIN_TYPE_IMPORT;
}

//---------------------------------------------------------------------------
//  MQEnumFileType
//    入力または出力可能なファイルタイプを返す。
//    ファイルタイプは別名保存時のダイアログに表示される。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileType(int index)
{
	switch(index){
	case 0: return "Metasequoia Document (*.mqo)";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQEnumFileExt
//    入力または出力可能な拡張子を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT const char *MQEnumFileExt(int index)
{
	switch(index){
	case 0: return "mqo";
	}
	return NULL;
}

//---------------------------------------------------------------------------
//  MQImportFile
//    [読み込み]やドラッグ＆ドロップで読み込むときに呼び出される。
//---------------------------------------------------------------------------
MQPLUGIN_EXPORT BOOL MQImportFile(int index, const char *filename, MQDocument doc)
{
	switch(index){
	case 0: return LoadMQO(filename, doc);
	}
	return FALSE;
}


// MQO形式を入力
BOOL LoadMQO(const char *filename, MQDocument doc)
{
	int ret_code;

	FILE *fh;
	errno_t ret_s = fopen_s(&fh, filename, "rb");
	if(ret_s != 0)
		return NULL;

	LoaderMQO *loader = new LoaderMQO(doc, fh);

	ret_code = loader->Load();

	delete loader;
	fclose(fh);

	return (ret_code == LOADERMQO_SUCCEEDED);
}



//---------------------------------------------------------------------------
//  LoaderMQO::LoaderMQO
//---------------------------------------------------------------------------
LoaderMQO::LoaderMQO(MQDocument document, FILE *handle)
{
	doc = document;

	fh = handle;
	afs = new AsciiFileString(fh);
}	//LoaderMQO::LoaderMQO()

//---------------------------------------------------------------------------
//  LoaderMQO::~LoaderMQO
//---------------------------------------------------------------------------
LoaderMQO::~LoaderMQO()
{
	delete afs;
}	//LoaderMQO::~LoaderMQO()

//---------------------------------------------------------------------------
//  LoaderMQO::CheckHeader
//---------------------------------------------------------------------------
bool LoaderMQO::CheckHeader(void)
{
	int sc;

	//check header
	sc = afs->GetLine();
	if(!afs->chkcom("Metasequoia",2) || !afs->chkstr(1,"Document"))
		throw(LOADERMQO_HEADER_ERR);

	//check format
	sc = afs->GetLine();
	if(!afs->chkcom("Format",4) || !afs->chkstr(1,"Text")
	|| !afs->chkstr(2,"Ver") || afs->ToDouble(3)!=1.0)
		throw(LOADERMQO_FORMAT_ERR);

	return true;
}	//LoaderMQO::CheckHeader()

//---------------------------------------------------------------------------
//  LoaderMQO::IgnoreBinary
//---------------------------------------------------------------------------
bool LoaderMQO::IgnoreBinary(int string_count)
{
	long oldpos, newpos;
	long offset = afs->ToInt(string_count-2);

	oldpos = ftell(fh);
	fseek(fh, offset, SEEK_CUR);
	newpos = ftell(fh);

	if(newpos - oldpos != offset)
		throw(LOADERMQO_ERROR);

	return true;
}	//LoaderMQO::IgnoreBinary()

//---------------------------------------------------------------------------
//  LoaderMQO::IgnoreUnknown
//---------------------------------------------------------------------------
bool LoaderMQO::IgnoreUnknown(int string_count)
{
	if(string_count >= 1 && CMPSTR(string_count-1,"{"))
	{   //pass unknown group
		for(int comc=1; comc>0; )
		{
			string_count = afs->GetLine();
			if(string_count == EOF)
			{
				throw(LOADERMQO_EOF);
			}
			else if(string_count >= 1)
			{
				if(CMPSTR(string_count-1,"{"))
				{
					comc++;
				}
				else if(CMPSTR(string_count-1,"}"))
				{
					comc--;
				}
				else if( string_count>=3
				&& afs->CheckString(string_count-3,"[")
				&& afs->CheckString(string_count-1,"]") )
				{
					IgnoreBinary(string_count-2);
				}
			}
		}
	}
	else if( string_count>=3 && afs->CheckString(string_count-3,"[") && afs->CheckString(string_count-1,"]") )
	{
		IgnoreBinary(string_count-2);
	}

	return true;
}	//LoaderMQO::IgnoreUnknown()

//---------------------------------------------------------------------------
//  LoaderMQO::MainLoop
//---------------------------------------------------------------------------
bool LoaderMQO::MainLoop(void)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF) break;
		if(sc == 0) continue;

		if( afs->chkcom("TrialNoise",2) )
		{
			throw(LOADERMQO_ERROR);
		}
		else if( afs->chkcom("Scene",2) )
		{
			if( !afs->chkstr(sc-1,"{") )
				throw(LOADERMQO_ERROR);
			if( !LoadScene() )
				throw(LOADERMQO_ERROR);
		}
		else if( afs->chkcom("Material",2) )
		{
			if( !afs->chkstr(sc-1,"{") )
				throw(LOADERMQO_ERROR);
			if( !LoadMaterial() )
				throw(LOADERMQO_ERROR);
		}
		else if( afs->chkcom("Object",2) )
		{
			if(!afs->chkstr(sc-1,"{"))
				throw(LOADERMQO_ERROR);

			MQObject curobj = MQ_CreateObject();
			int result = 0;
			try {
				if(sc >= 3) {
					curobj->SetName(afs->GetString(1));
				}
				if( !LoadObject(curobj) )
					throw(LOADERMQO_ERROR);
				// TODO: 親子関係の構築は行われません。
				doc->AddObject(curobj);
			}
			catch(int err_code)
			{
				curobj->DeleteThis();
				result = err_code;
			}
			if(result)
				throw(result);
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}

	return true;
}	//LoaderMQO::MainLoop()

//---------------------------------------------------------------------------
//  LoaderMQO::Load
//---------------------------------------------------------------------------
int LoaderMQO::Load(void)
{
	try
	{
		if( !CheckHeader() )
			throw(LOADERMQO_ERROR);
		if( !MainLoop() )
			throw(LOADERMQO_ERROR);
	}
	catch(int err_code)
	{
		return err_code;
	}

	return LOADERMQO_SUCCEEDED;
}	//LoaderMQO::Load()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadScene
//---------------------------------------------------------------------------
bool LoaderMQO::LoadScene(void)
{
	MQScene scene = doc->GetScene(0);
	if (scene == NULL)
		return false;

	MQAngle angle(0,0,0);
	bool angle_valid = false;

	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);

		if(COMCHECK(1,"}"))
		{
			break;
		}
		/*else if(COMCHECK(4,"pos"))
		{
		}*/
#if MQPLUGIN_VERSION >= 0x0240
		else if(COMCHECK(4,"lookat"))
		{
			scene->SetLookAtPosition( MQPoint(afs->ToFloat(1),afs->ToFloat(2),afs->ToFloat(3)) );
		}
		else if(COMCHECK(2,"head"))
		{
			angle.head = afs->ToFloat(1);
			angle_valid = true;
		}
		else if(COMCHECK(2,"pich"))
		{
			angle.pich = afs->ToFloat(1);
			angle_valid = true;
		}
		else if(COMCHECK(2,"bank"))
		{
			angle.bank = afs->ToFloat(1);
			angle_valid = true;
		}
#endif
		/*else if(COMCHECK(2,"ortho"))
		{
			scene->ortho = (MBool)afs->ToInt(1);
		}
		else if(COMCHECK(2,"zoom2"))
		{
			scene->zoom = afs->ToFloat(1);
		}*/
		else if(COMCHECK(4,"amb"))
		{
			scene->SetGlobalAmbientColor( MQColor(afs->ToFloat(1),afs->ToFloat(2),afs->ToFloat(3)) );
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}

	if (angle_valid){
		scene->SetCameraAngle(angle);
	}

	return true;
}	//LoaderMQO::LoadScene()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadMaterial
//---------------------------------------------------------------------------
bool LoaderMQO::LoadMaterial(void)
{
	int i,scn;
	MQMaterial mat;

	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(sc == 0) continue;

		if(COMCHECK(1,"}")) break;

		mat = MQ_CreateMaterial();

		mat->SetName(afs->GetString(0));

		for(scn=1; scn<sc;)
		{
			if(CMPSTR(scn,"col") && scn+7<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+6,")")) break;
				mat->SetColor( MQColor(
					afs->ToFloat(scn+2), afs->ToFloat(scn+3), afs->ToFloat(scn+4)) );
				mat->SetAlpha(afs->ToFloat(scn+5));
				scn += 7;
			}
			else if(CMPSTR(scn,"dif") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetDiffuse(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"amb") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetAmbient(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"spc") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetSpecular(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"emi") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetEmission(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"power") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetPower(afs->ToFloat(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"tex") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetTextureName(afs->GetString(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"aplane") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetAlphaName(afs->GetString(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"bump") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetBumpName(afs->GetString(scn+2));
				scn += 4;
			}
#if MQPLUGIN_VERSION >= 0x0230
			else if(CMPSTR(scn,"shader") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetShader( afs->ToInt(scn+2) );
				scn += 4;
			}
			else if(CMPSTR(scn,"vcol") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetVertexColor( afs->ToInt(scn+2) );
				scn += 4;
			}
			else if(CMPSTR(scn,"proj_type") && scn+4<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) break;
				mat->SetMappingType(afs->ToInt(scn+2));
				scn += 4;
			}
			else if(CMPSTR(scn,"proj_pos") && scn+6<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+5,")")) break;
				mat->SetMappingPosition(MQPoint(
					afs->ToFloat(scn+2), afs->ToFloat(scn+3), afs->ToFloat(scn+4) ));
				scn += 6;
			}
			else if(CMPSTR(scn,"proj_scale") && scn+6<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+5,")")) break;
				mat->SetMappingScaling(MQPoint(
					afs->ToFloat(scn+2), afs->ToFloat(scn+3), afs->ToFloat(scn+4) ));
				scn += 6;
			}
			else if(CMPSTR(scn,"proj_angle") && scn+6<=sc)
			{
				if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+5,")")) break;
				mat->SetMappingAngle(MQAngle(
					afs->ToFloat(scn+2), afs->ToFloat(scn+3), afs->ToFloat(scn+4) ));
				scn += 6;
			}
#endif
			else if(CMPSTR(scn+1,"("))
			{
				for(i=2; i+scn<sc; i++) {
					if(CMPSTR(scn+i,")")) {
						++i;
						break;
					}
				}
				scn += i;
			}
			else
			{
				break;
			}
		}

		doc->AddMaterial(mat);
	}

	return true;
}	//LoaderMQO::LoadMaterial()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadObject
//---------------------------------------------------------------------------
bool LoaderMQO::LoadObject(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);

		if(COMCHECK(1,"}"))
		{
			break;
		}
		/*else if(COMCHECK(2,"depth"))
		{
			obj->SetDepth( afs->ToInt(1) );
		}
		else if(COMCHECK(2,"folding"))
		{
			obj->SetFolding( afs->ToInt(1) );
		}*/
#if MQPLUGIN_VERSION >= 0x0240
		else if(COMCHECK(2,"scale"))
		{
			obj->SetScaling( MQPoint(afs->ToFloat(1),afs->ToFloat(2),afs->ToFloat(3)) );
		}
		else if(COMCHECK(2,"rotation"))
		{
			obj->SetRotation( MQAngle(afs->ToFloat(1),afs->ToFloat(2),afs->ToFloat(3)) );
		}
		else if(COMCHECK(2,"translation"))
		{
			obj->SetTranslation( MQPoint(afs->ToFloat(1),afs->ToFloat(2),afs->ToFloat(3)) );
		}
#endif
		else if(COMCHECK(2,"patch"))
		{
			obj->SetPatchType(afs->ToInt(1));
		}
		else if(COMCHECK(2,"segment"))
		{
			obj->SetPatchSegment(afs->ToInt(1));
		}
		else if(COMCHECK(2,"visible"))
		{
			obj->SetVisible(afs->ToInt(1));
		}
		/*else if(COMCHECK(2,"locking"))
		{
			obj->locking = (MBool)afs->ToInt(1);
		}*/
		else if(COMCHECK(2,"shading"))
		{
			obj->SetShading(afs->ToInt(1));
		}
		else if(COMCHECK(2,"facet"))
		{
			obj->SetSmoothAngle(afs->ToFloat(1));
		}
		/*else if(COMCHECK(4,"color"))
		{
			obj->SetPreviewColor(MQColor(
				afs->ToFloat(1), afs->ToFloat(2), afs->ToFloat(3) ));
		}
		else if(COMCHECK(2,"color_type"))
		{
			obj->SetColorType( afs->ToInt(1) );
		}*/
		else if(COMCHECK(2,"mirror"))
		{
			obj->SetMirrorType(afs->ToInt(1));
		}
		else if(COMCHECK(2,"mirror_axis"))
		{
			obj->SetMirrorAxis(afs->ToInt(1));
		}
		else if(COMCHECK(2,"mirror_dis"))
		{
			obj->SetMirrorDistance(afs->ToFloat(1));
		}
		else if(COMCHECK(2,"lathe"))
		{
			obj->SetLatheType(afs->ToInt(1));
		}
		else if(COMCHECK(2,"lathe_axis"))
		{
			obj->SetLatheAxis(afs->ToInt(1));
		}
		else if(COMCHECK(2,"lathe_seg"))
		{
			obj->SetLatheSegment(afs->ToInt(1));
		}
		else if(COMCHECK(3,"vertex"))
		{
			if(!CMPSTR(sc-1,"{")) return false;
			if(!LoadVertex(obj)) return false;
		}
		else if(COMCHECK(2,"vertexattr"))
		{
			if(!CMPSTR(sc-1,"{")) return false;
			if(!LoadVertexAttr(obj)) return false;
		}
		else if(COMCHECK(3,"BVertex"))
		{
			if(!CMPSTR(sc-1,"{")) return false;
			if(!LoadBVertex(obj)) return false;
		}
		else if(COMCHECK(2,"face"))
		{
			if(!CMPSTR(sc-1,"{")) return false;
			if(!LoadFace(obj)) return false;
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}

	return true;
}	//LoaderMQO::LoadObject()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadVertex
//---------------------------------------------------------------------------
bool LoaderMQO::LoadVertex(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(COMCHECK(1,"}")) break;

		if(sc < 3) continue;

		obj->AddVertex(MQPoint(
			afs->ToFloat(0), afs->ToFloat(1), afs->ToFloat(2) ));
	}

	return true;
}	//LoaderMQO::LoadVertex()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadVertexAttr
//---------------------------------------------------------------------------
bool LoaderMQO::LoadVertexAttr(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);

		if(COMCHECK(1,"}"))
		{
			break;
		}
		else if(COMCHECK(2,"color"))
		{
			LoadVertexAttrColor(obj);
		}
		else if(COMCHECK(2,"weit"))
		{
			LoadVertexAttrWeight(obj);
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}

	return true;
}	//LoaderMQO::LoadVertexAttr()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadVertexAttrColor
//---------------------------------------------------------------------------
bool LoaderMQO::LoadVertexAttrColor(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(COMCHECK(1,"}")) break;

		if(sc < 2) continue;

		int vertIdx = afs->ToInt(0);
		if( vertIdx >= 0 && vertIdx < obj->GetVertexCount() )
		{
			//obj->SetVertexColor( vertIdx, afs->ToInt(1) ); // OBSOLUTE
		}
	}

	return true;
}	//LoaderMQO::LoadVertexAttrColor()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadVertexAttrWeight
//---------------------------------------------------------------------------
bool LoaderMQO::LoadVertexAttrWeight(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(COMCHECK(1,"}")) break;

		if(sc < 2) continue;

		int vertIdx = afs->ToInt(0);
		if( vertIdx >= 0 && vertIdx < obj->GetVertexCount() )
		{
			obj->SetVertexWeight( vertIdx, afs->ToFloat(1) );
		}
	}

	return true;
}	//LoaderMQO::LoadVertexAttrWeight()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadBVertex
//---------------------------------------------------------------------------
bool LoaderMQO::LoadBVertex(MQObject obj)
{
	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(COMCHECK(1,"}")) break;

		if(COMCHECK(5,"Vector"))
		{
			if(!CMPSTR(sc-3,"[") || !CMPSTR(sc-1,"]")) return false;
			int vc = afs->ToInt(1);

			float vec[3];
			for(int i=0; i<vc; i++) {
				fread(vec, 1, sizeof(vec), afs->GetFileHandle());
				obj->AddVertex(MQPoint(vec[0], vec[1], vec[2]));
			}
		}
		else if(COMCHECK(2,"color"))
		{
			for(;;)
			{
				sc = afs->GetLine();
				if(sc == EOF) break;
				else if(CMPSTR(sc-1,"}")) break;
				int vi = afs->ToInt(0);
				if(vi < obj->GetVertexCount()){
					//obj->SetVertexColor(vi, (DWORD)afs->ToInt(1)); //OBSOLUTE
				}
			}
		}
		else if(COMCHECK(2,"weit"))
		{
			for(;;)
			{
				sc = afs->GetLine();
				if(sc == EOF) break;
				else if(CMPSTR(sc-1,"}")) break;
				int vi = afs->ToInt(0);
				if(vi < obj->GetVertexCount())
					obj->SetVertexWeight(vi, afs->ToFloat(1));
			}
		}
		else
		{
			IgnoreUnknown(sc);
		}
	}
	return true;
}	//LoaderMQO::LoadBVertex()

//---------------------------------------------------------------------------
//  LoaderMQO::LoadFace
//---------------------------------------------------------------------------
bool LoaderMQO::LoadFace(MQObject obj)
{
	int scn;
	int vc,cfi;

	for(;;)
	{
		int sc = afs->GetLine();
		if(sc == EOF)
			throw(LOADERMQO_EOF);
		if(sc == 0) continue;

		if(COMCHECK(1,"}")) break;

		vc = afs->ToInt(0); //vertex count
		scn = 1;
		if(CMPSTR(scn,"V") && scn+vc+3<=sc)
		{
			int vert_index[4];
			if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+vc+2,")")) continue;
			if(vc>=1) vert_index[0] = afs->ToInt(scn+2);
			if(vc>=2) vert_index[1] = afs->ToInt(scn+3);
			if(vc>=3) vert_index[2] = afs->ToInt(scn+4);
			if(vc>=4) vert_index[3] = afs->ToInt(scn+5);
			cfi = obj->AddFace(vc, vert_index);
			scn += vc+3;
		} else {
			throw(LOADERMQO_ERROR);
		}

		if(CMPSTR(scn,"M") && scn+4<=sc)
		{
			if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+3,")")) continue;
			int pal = afs->ToInt(scn+2);
			if(0<=pal && pal<doc->GetMaterialCount())
				obj->SetFaceMaterial(cfi, pal);
			scn += 4;
		}

		if(CMPSTR(scn,"UV") && scn+vc*2+3<=sc)
		{
			MQCoordinate uv[4];

			if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+vc*2+2,")")) continue;
			
			if(vc>=1) { uv[0].u=afs->ToFloat(scn+2); uv[0].v=afs->ToFloat(scn+3); }
			if(vc>=2) { uv[1].u=afs->ToFloat(scn+4); uv[1].v=afs->ToFloat(scn+5); }
			if(vc>=3) { uv[2].u=afs->ToFloat(scn+6); uv[2].v=afs->ToFloat(scn+7); }
			if(vc>=4) { uv[3].u=afs->ToFloat(scn+8); uv[3].v=afs->ToFloat(scn+9); }

			obj->SetFaceCoordinateArray(cfi, uv);

			scn += vc*2+3;
		}

#if MQPLUGIN_VERSION >= 0x0230
		if(CMPSTR(scn,"COL") && scn+vc+3<=sc)
		{
			if(!CMPSTR(scn+1,"(") || !CMPSTR(scn+vc+2,")")) continue;
			if(vc>=1) obj->SetFaceVertexColor(cfi, 0, afs->ToInt(scn+2));
			if(vc>=2) obj->SetFaceVertexColor(cfi, 1, afs->ToInt(scn+3));
			if(vc>=3) obj->SetFaceVertexColor(cfi, 2, afs->ToInt(scn+4));
			if(vc>=4) obj->SetFaceVertexColor(cfi, 3, afs->ToInt(scn+5));
			scn += vc+3;
		}
#endif

	}

	return true;
}	//LoaderMQO::LoadFace()

