﻿//---------------------------------------------------------------------------
//
//   Scissor
//
//          Copyright(C) 2007-2012, O.Mizno
//
//     Sample of Command plug-in for a knife.
//     Created DLL must be installed in "Plugins\Command" directory. 
//     After installing the plug-in, the "Scissor" button will be shown 
//    in the command panel.
//
//    　ナイフのコマンドプラグインのサンプル。
//    　作成したDLLは"Plugins\Command"フォルダに入れる必要がある。
//    　プラグインをインストールすると"Scissor"ボタンが
//    コマンドパネルに追加されます。
//
//---------------------------------------------------------------------------
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <float.h>
#include <vector>
#include "MQBasePlugin.h"
#include "MQ3DLib.h"


struct MQRGBAColor
{
	float r,g,b,a;
	MQRGBAColor() {}
	MQRGBAColor(float _r,float _g,float _b,float _a) {r=_r; g=_g; b=_b; a=_a;}
	MQRGBAColor(const MQRGBAColor& color) {r=color.r; g=color.g; b=color.b; a=color.a;}
	MQRGBAColor(DWORD color) {
		r=(float)( color&0x000000FF)/255;
		g=(float)((color&0x0000FF00)>>8)/255;
		b=(float)((color&0x00FF0000)>>16)/255;
		a=(float)((color&0xFF000000)>>24)/255;
	}

	friend inline MQRGBAColor operator + (const MQRGBAColor& c1, const MQRGBAColor& c2)
		{return MQRGBAColor(c1.r+c2.r, c1.g+c2.g, c1.b+c2.b, c1.a+c2.a);}
	friend inline MQRGBAColor operator * (const MQRGBAColor& c, const float& s)
		{return MQRGBAColor(c.r*s, c.g*s, c.b*s, c.a*s);}

	DWORD GetRGBA() const {
		return (BYTE)(r*255.0f+0.5f) | 
			  ((BYTE)(g*255.0f+0.5f) << 8) |
			  ((BYTE)(b*255.0f+0.5f) << 16) |
			  ((BYTE)(a*255.0f+0.5f) << 24);
		}
};


//---------------------------------------------------------------------------
//  IsCrossing2D
//---------------------------------------------------------------------------
static bool IsCrossing2D(
	const MQPoint& s0, const MQPoint& s1,
	const MQPoint& t0, const MQPoint& t1,
	float *outs, float *outt)
{
	float f,s,t;

	f = (s1.x-s0.x)*(t1.y-t0.y) - (s1.y-s0.y)*(t1.x-t0.x);
	if(f == 0) return false;
	s = ((s0.y-t0.y)*(t1.x-t0.x) - (s0.x-t0.x)*(t1.y-t0.y)) / f;
	t = ((s0.y-t0.y)*(s1.x-s0.x) - (s0.x-t0.x)*(s1.y-s0.y)) / f;
	if(0<=t && t<1 && 0<=s && s<1){
		if(outs != NULL) *outs = s;
		if(outt != NULL) *outt = t;
		return true;
	}

	return false;
}



class ScissorPlugin : public MQCommandPlugin
{
public:
	// コンストラクタ
	ScissorPlugin();

	// プラグインIDを返す。
	virtual void GetPlugInID(DWORD *Product, DWORD *ID);
	// プラグイン名を返す。
	virtual const char *GetPlugInName(void);
	// ポップアップメニューに表示される文字列を返す。
	virtual const char *EnumString(void);
	// アプリケーションの初期化
	virtual BOOL Initialize();
	// アプリケーションの終了
	virtual void Exit();
	// 表示・非表示切り替え要求
	virtual BOOL Activate(MQDocument doc, BOOL flag);
	// 描画時の処理
	virtual void OnDraw(MQDocument doc, MQScene scene, int width, int height);

	// 左ボタンが押されたとき
	virtual BOOL OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	// 左ボタンが押されながらマウスが移動したとき
	virtual BOOL OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);
	// 左ボタンが離されたとき
	virtual BOOL OnLeftButtonUp(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state);

private:
	bool m_bDragging;
	POINT m_KnifePos[2];	// ナイフの始点・終点

	// ナイフ処理の実行
	void ExecKnife(MQDocument doc, MQScene scene);

};



//---------------------------------------------------------------------------
//  ScissorPlugin::ScissorPlugin
//    コンストラクタ
//---------------------------------------------------------------------------
ScissorPlugin::ScissorPlugin()
{
	m_bDragging = false;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::GetPlugInID
//    プラグインIDを返す。
//---------------------------------------------------------------------------
void ScissorPlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x01AE46E3;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::GetPlugInName
//    プラグイン名を返す。
//---------------------------------------------------------------------------
const char *ScissorPlugin::GetPlugInName(void)
{
	return "Scissor       Copyright(C) 2007-2012, O.Mizno";
}


//---------------------------------------------------------------------------
//  ScissorPlugin::EnumString
//    ボタンに表示される文字列を返す。
//---------------------------------------------------------------------------
const char *ScissorPlugin::EnumString(void)
{
	return "Scissor";
}


//---------------------------------------------------------------------------
//  ScissorPlugin::Initialize
//    アプリケーションの初期化
//---------------------------------------------------------------------------
BOOL ScissorPlugin::Initialize()
{
	// 特に何もしないので、そのままTRUEを返す
	return TRUE;
}

//---------------------------------------------------------------------------
//  ScissorPlugin::Exit
//    アプリケーションの終了
//---------------------------------------------------------------------------
void ScissorPlugin::Exit()
{
	// 特に何もしない
}

//---------------------------------------------------------------------------
//  ScissorPlugin::Activate
//    表示・非表示切り替え要求
//---------------------------------------------------------------------------
BOOL ScissorPlugin::Activate(MQDocument doc, BOOL flag)
{
	if (!flag){
		if(m_bDragging){
			m_bDragging = false;
			RedrawAllScene();
		}
	}
	
	// そのままflagを返す
	return flag;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::OnDraw
//    描画時の処理
//---------------------------------------------------------------------------
void ScissorPlugin::OnDraw(MQDocument doc, MQScene scene, int width, int height)
{
	// ナイフの始点・終点を描画
	if (m_bDragging)
	{
		// 頂点位置を計算
		MQPoint v[2];
		v[0] = scene->ConvertScreenTo3D(MQPoint((float)m_KnifePos[0].x,(float)m_KnifePos[0].y,0.00001f)); // 手前に描画するためZにほぼ0（計算誤差を防ぐため）をセット
		v[1] = scene->ConvertScreenTo3D(MQPoint((float)m_KnifePos[1].x,(float)m_KnifePos[1].y,0.00001f)); // 手前に描画するためZにほぼ0（計算誤差を防ぐため）をセット

		// 描画オブジェクトを生成
		int vertex[2];
		MQObject draw = CreateDrawingObject(doc, DRAW_OBJECT_LINE);
		draw->SetColor(MQColor(1,1,0));
		draw->SetColorValid(TRUE);
		vertex[0] = draw->AddVertex(v[0]);
		vertex[1] = draw->AddVertex(v[1]);
		draw->AddFace(2, vertex);
	}
}


//---------------------------------------------------------------------------
//  ScissorPlugin::OnLeftButtonDown
//    左ボタンが押されたとき
//---------------------------------------------------------------------------
BOOL ScissorPlugin::OnLeftButtonDown(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	// クリック位置を記録しておく
	m_KnifePos[0] = state.MousePos;
	m_KnifePos[1] = state.MousePos;
	m_bDragging = true;

	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::OnLeftButtonMove
//    左ボタンが押されながらマウスが移動したとき
//---------------------------------------------------------------------------
BOOL ScissorPlugin::OnLeftButtonMove(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (m_bDragging)
	{
		// 始点とカーソル位置を結ぶ
		if (state.Shift){
			if (abs(state.MousePos.x-m_KnifePos[0].x) >= abs(state.MousePos.y-m_KnifePos[0].y)){
				m_KnifePos[1].x = state.MousePos.x;
				m_KnifePos[1].y = m_KnifePos[0].y;
			}else{
				m_KnifePos[1].x = m_KnifePos[0].x;
				m_KnifePos[1].y = state.MousePos.y;
			}
		}else{
			m_KnifePos[1] = state.MousePos;
		}

		// 再描画
		RedrawScene(scene);
	}

	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


//---------------------------------------------------------------------------
//  ScissorPlugin::OnLeftButtonUp
//    左ボタンが離されたとき
//---------------------------------------------------------------------------
BOOL ScissorPlugin::OnLeftButtonUp(MQDocument doc, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if (m_bDragging)
	{
		m_bDragging = false;

		if (m_KnifePos[0].x != m_KnifePos[1].x || m_KnifePos[0].y != m_KnifePos[1].y){
			// ナイフの実行
			ExecKnife(doc, scene);

			// 頂点が移動された場合、再描画とアンドゥの更新を行う
			UpdateUndo();
		}

		RedrawAllScene();
	}

	// 標準動作の代わりに独自処理を行ったのでTRUEを返す
	return TRUE;
}


struct KnifeHash {
	KnifeHash *next;
	int v2;
	int nvi, newobj_nvi; // 新規に出来た頂点のインデックス
	float t;
	MQPoint v;
};

void ScissorPlugin::ExecKnife(MQDocument doc, MQScene scene)
{
	char buf[256];
	int j,k;

	// 編集オプションを取得
	EDIT_OPTION option;
	GetEditOption(option);

	MQPoint tsv1, tsv2;
	tsv1.x = (float)m_KnifePos[0].x;
	tsv1.y = (float)m_KnifePos[0].y;
	tsv1.z = 0;
	tsv2.x = (float)m_KnifePos[1].x;
	tsv2.y = (float)m_KnifePos[1].y;
	tsv2.z = 0;

	// 切断面の法線を求める
	MQPoint plane1 = scene->ConvertScreenTo3D(tsv1);
	MQPoint plane2 = scene->ConvertScreenTo3D(tsv2);
	MQPoint plane3 = scene->ConvertScreenTo3D(MQPoint(tsv1.x,tsv1.y,0.25f));
	MQPoint normal = GetNormal(plane1, plane2, plane3);

	for(int oi=0; oi<doc->GetObjectCount(); oi++)
	{
		MQObject obj = doc->GetObject(oi);
		if (obj == NULL) continue;

		// 非表示またはロックされていれば対象外
		if (!obj->GetVisible() || obj->GetLocking()) continue;

		// カレントオブジェクトのみ編集
		if (option.CurrentObjectOnly && oi != doc->GetCurrentObjectIndex())
			continue;

		int fc = obj->GetFaceCount();
		int vc = obj->GetVertexCount();

		// 面が見えるかどうか調べておく
		BOOL *visible = new BOOL[fc];
		scene->GetVisibleFace(obj, visible);

		bool *cut_face = new bool[fc];
		for(j=0; j<fc; j++)
			cut_face[j] = false;

		// allocate hash
		KnifeHash *hash, *ch;
		int hashsize = fc*4 + vc;
		hash = (KnifeHash *)malloc(sizeof(KnifeHash) * hashsize);
		for(j=0; j<hashsize; j++){
			hash[j].next = NULL;
			hash[j].v2 = hash[j].nvi = hash[j].newobj_nvi = -1;
		}
		int hc = obj->GetVertexCount(); //count of current hash

		// 交差しているラインを検知
		int count = 0;
		for(j=0; j<fc; j++)
		{
			int cf_count = obj->GetFacePointCount(j);
			if(cf_count == 0) continue;

			// 見えない面は対象外
			if(!visible[j]) continue;

			int vert[4];
			obj->GetFacePointArray(j, vert);

			for(k=0; k<cf_count; k++)
			{
				if(k == 1 && cf_count == 2)
					break;

				// check hash
				int fv1=vert[k], fv2=vert[(k+1)%cf_count];
				int hv1 = min(fv1,fv2);
				int hv2 = max(fv1,fv2);
				bool hash_found = false;
				for(ch=&hash[hv1]; ch!=NULL && ch->v2>=0; ch=ch->next){
					if(ch->v2 == hv2){
						hash_found = true;
						break;
					}
				}
				if (hash_found)
					continue;

				// check if crossing
				MQPoint screen1,screen2;
				float w1,w2;
				screen1 = scene->Convert3DToScreen(obj->GetVertex(hv1),&w1);
				screen2 = scene->Convert3DToScreen(obj->GetVertex(hv2),&w2);
				float t;
				if(!IsCrossing2D(tsv1, tsv2, screen1, screen2, NULL, &t))
					continue;

				// add to hash
				for(ch=&hash[hv1]; ; ch=ch->next)
				{
					if(ch->v2 < 0)
					{
						ch->v2 = hv2;

						MQPoint sv0,v0;

						sv0 = screen1*(1-t) + screen2*t;
						v0 = scene->ConvertScreenTo3D(sv0);

						float d1 = GetSize(v0 - obj->GetVertex(hv1));
						float d2 = GetSize(obj->GetVertex(hv2) - obj->GetVertex(hv1));
						if(d2 > 1e-8f)
							ch->t = d1 / d2;
						else
							ch->t = t;
						ch->v = obj->GetVertex(hv1)*(1-ch->t) + obj->GetVertex(hv2)*ch->t;
						break;
					}
					if(ch->next == NULL)
						ch->next = &hash[hc++];
				}

				cut_face[j] = true;
				count++;
			}
		}

		// 切断されるラインがなければパス
		if (count == 0){
			free(hash);
			delete[] cut_face;
			delete[] visible;
			continue;
		}

		// オブジェクト生成
		MQObject newobj = MQ_CreateObject();
		{
			// 名前を付ける
			char name[64], newname[64];
			obj->GetName(name, sizeof(name));

			// 重複しない連番オブジェクト名を付ける
			for (int inc=1; inc<1000; inc++){
				sprintf_s(newname, sizeof(newname), "%s%d", name, inc);
				bool exist = false;
				for (int n=0; n<doc->GetObjectCount(); n++){
					if (n == oi) continue;
					
					MQObject obj = doc->GetObject(n);
					if (obj == NULL) continue;

					char tempname[64];
					obj->GetName(tempname, sizeof(tempname));
					if (strcmp(tempname, newname) == 0){
						exist = true;
						break;
					}
				}

				if (!exist){
					newobj->SetName(newname);
					break;
				}
			}
		}

		// 新規オブジェクトを同じ階層の下側に挿入
		doc->InsertObject(newobj, obj);
		newobj->SetDepth(obj->GetDepth());

		// 切断面を中心に頂点を分離
		int *newobj_vi = new int[vc];
		count = 0;
		for(j=0; j<vc; j++){
			newobj_vi[j] = -1;
			if(obj->GetVertexRefCount(j) > 0){
				MQPoint v = obj->GetVertex(j);
				if (GetInnerProduct(v-plane1, normal) > 0){
					newobj_vi[j] = newobj->AddVertex(v);
					count++;
				}
			}
		}
		sprintf_s(buf, sizeof(buf), "Scissor: separated %d vertices\n", count);
		OutputDebugString(buf);

		// 切断ライン上の頂点を追加
		count = 0;
		for(j=0,ch=hash; j<hc; j++,ch++)
		{
			if(ch->v2 >= 0){
				ch->nvi = obj->AddVertex(ch->v);
				ch->newobj_nvi = newobj->AddVertex(ch->v);
				count++;
			}		
		}
		sprintf_s(buf, sizeof(buf), "Scissor: divide %d lines\n", count);
		OutputDebugString(buf);

		// 切断面の構築
		for(j=0; j<fc; j++)
		{
			int hv1,hv2,nfi;
			int nvi[4], newobj_nvi[4];
			int nfvert[4];
			MQCoordinate nfcoord[4];
			float t[4];

			int cf_count = obj->GetFacePointCount(j);
			if(obj->GetFacePointCount(j) == 0) continue;

			int cfvi[4];
			MQCoordinate cfcoord[4];
			DWORD cfcolor[4];
			obj->GetFacePointArray(j, cfvi);
			obj->GetFaceCoordinateArray(j, cfcoord);
			for(k=0; k<cf_count; k++)
				cfcolor[k] = obj->GetFaceVertexColor(j,k);

			int pal = obj->GetFaceMaterial(j);

			// 分断されるライン上の頂点インデックスをnviにセット
			for(k=0; k<cf_count; k++)
			{
				nvi[k] = newobj_nvi[k] = -1;
				// search hash
				hv1 = min(cfvi[k], cfvi[(k+1)%cf_count]);
				hv2 = max(cfvi[k], cfvi[(k+1)%cf_count]);
				for(ch=&hash[hv1]; ch!=NULL && ch->v2>=0; ch=ch->next){
					if(ch->v2 == hv2){
						nvi[k] = ch->nvi;
						newobj_nvi[k] = ch->newobj_nvi;
						t[k] = ch->t;
						break;
					}
				}
			}

			// 面・ラインを切断
			bool divided = false;
			switch(cf_count) {
			case 2:
				if(nvi[0]>=0)
				{
					if (newobj_vi[cfvi[0]] != -1){
						nfvert[0] = newobj_vi[cfvi[0]];
						nfvert[1] = newobj_nvi[0];
						nfi = newobj->AddFace(2, nfvert);
						newobj->SetFaceMaterial(nfi, pal);
					}else{
						nfvert[0] = cfvi[0];
						nfvert[1] = nvi[0];
						nfi = obj->AddFace(2, nfvert);
						obj->SetFaceMaterial(nfi, pal);
					}

					if (newobj_vi[cfvi[1]] != -1){
						nfvert[0] = newobj_vi[cfvi[1]];
						nfvert[1] = newobj_nvi[0];
						nfi = newobj->AddFace(2, nfvert);
						newobj->SetFaceMaterial(nfi, pal);
					}else{
						nfvert[0] = cfvi[1];
						nfvert[1] = nvi[0];
						nfi = obj->AddFace(2, nfvert);
						obj->SetFaceMaterial(nfi, pal);
					}

					obj->DeleteFace(j);
					divided = true;
				}
				break;

			case 3:
				for(k=0; k<3; k++)
				{
					if(nvi[k]>=0 && nvi[(k+1)%3]>=0)
					{
						MQCoordinate t0,t1;
						DWORD col0,col1;
						if(cfvi[k] < cfvi[(k+1)%3]){
							t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%3]*t[k];
							col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%3])*t[k]).GetRGBA();
						}else{
							t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%3]*(1-t[k]);
							col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%3])*(1-t[k])).GetRGBA();
						}
						if(cfvi[(k+1)%3] < cfvi[(k+2)%3]){
							t1 = cfcoord[(k+1)%3]*(1-t[(k+1)%3]) + cfcoord[(k+2)%3]*t[(k+1)%3];
							col1 = (MQRGBAColor(cfcolor[(k+1)%3])*(1-t[(k+1)%3]) + MQRGBAColor(cfcolor[(k+2)%3])*t[(k+1)%3]).GetRGBA();
						}else{
							t1 = cfcoord[(k+1)%3]*t[(k+1)%3] + cfcoord[(k+2)%3]*(1-t[(k+1)%3]);
							col1 = (MQRGBAColor(cfcolor[(k+1)%3])*t[(k+1)%3] + MQRGBAColor(cfcolor[(k+2)%3])*(1-t[(k+1)%3])).GetRGBA();
						}

						if (newobj_vi[cfvi[(k+1)%3]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_vi[cfvi[(k+1)%3]];
							nfvert[2] = newobj_nvi[(k+1)%3];
							nfi = newobj->AddFace(3, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%3];
							nfcoord[2] = t1;
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%3]);
							newobj->SetFaceVertexColor(nfi,2, col1);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = cfvi[(k+1)%3];
							nfvert[2] = nvi[(k+1)%3];
							nfi = obj->AddFace(3, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%3];
							nfcoord[2] = t1;
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%3]);
							obj->SetFaceVertexColor(nfi,2, col1);
						}

						if (newobj_vi[cfvi[k]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_nvi[(k+1)%3];
							nfvert[2] = newobj_vi[cfvi[(k+2)%3]];
							nfvert[3] = newobj_vi[cfvi[k]];
							nfi = newobj->AddFace(4, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+2)%3];
							nfcoord[3] = cfcoord[k];
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, col1);
							newobj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%3]);
							newobj->SetFaceVertexColor(nfi,3, cfcolor[k]);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = nvi[(k+1)%3];
							nfvert[2] = cfvi[(k+2)%3];
							nfvert[3] = cfvi[k];
							nfi = obj->AddFace(4, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+2)%3];
							nfcoord[3] = cfcoord[k];
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, col1);
							obj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%3]);
							obj->SetFaceVertexColor(nfi,3, cfcolor[k]);
						}

						obj->DeleteFace(j);
						divided = true;
						break;
					}
				}
				break;

			case 4:
				for(k=0; k<4; k++)
				{
					if(nvi[k]>=0 && nvi[(k+2)%4]>=0)
					{
						MQCoordinate t0,t1;
						DWORD col0,col1;
						if(cfvi[k] < cfvi[(k+1)%4]){
							t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
							col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
						}else{
							t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
							col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
						}
						if(cfvi[(k+2)%4] < cfvi[(k+3)%4]){
							t1 = cfcoord[(k+2)%4]*(1-t[(k+2)%4]) + cfcoord[(k+3)%4]*t[(k+2)%4];
							col1 = (MQRGBAColor(cfcolor[(k+2)%4])*(1-t[(k+2)%4]) + MQRGBAColor(cfcolor[(k+3)%4])*t[(k+2)%4]).GetRGBA();
						}else{
							t1 = cfcoord[(k+2)%4]*t[(k+2)%4] + cfcoord[(k+3)%4]*(1-t[(k+2)%4]);
							col1 = (MQRGBAColor(cfcolor[(k+2)%4])*t[(k+2)%4] + MQRGBAColor(cfcolor[(k+3)%4])*(1-t[(k+2)%4])).GetRGBA();
						}

						if (newobj_vi[cfvi[(k+1)%4]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_vi[cfvi[(k+1)%4]];
							nfvert[2] = newobj_vi[cfvi[(k+2)%4]];
							nfvert[3] = newobj_nvi[(k+2)%4];
							nfi = newobj->AddFace(4, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%4];
							nfcoord[2] = cfcoord[(k+2)%4];
							nfcoord[3] = t1;
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%4]);
							newobj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%4]);
							newobj->SetFaceVertexColor(nfi,3, col1);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = cfvi[(k+1)%4];
							nfvert[2] = cfvi[(k+2)%4];
							nfvert[3] = nvi[(k+2)%4];
							nfi = obj->AddFace(4, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%4];
							nfcoord[2] = cfcoord[(k+2)%4];
							nfcoord[3] = t1;
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%4]);
							obj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%4]);
							obj->SetFaceVertexColor(nfi,3, col1);
						}

						if (newobj_vi[cfvi[k]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_nvi[(k+2)%4];
							nfvert[2] = newobj_vi[cfvi[(k+3)%4]];
							nfvert[3] = newobj_vi[cfvi[k]];
							nfi = newobj->AddFace(4, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+3)%4];
							nfcoord[3] = cfcoord[k];
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, col1);
							newobj->SetFaceVertexColor(nfi,2, cfcolor[(k+3)%4]);
							newobj->SetFaceVertexColor(nfi,3, cfcolor[k]);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = nvi[(k+2)%4];
							nfvert[2] = cfvi[(k+3)%4];
							nfvert[3] = cfvi[k];
							nfi = obj->AddFace(4, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+3)%4];
							nfcoord[3] = cfcoord[k];
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, col1);
							obj->SetFaceVertexColor(nfi,2, cfcolor[(k+3)%4]);
							obj->SetFaceVertexColor(nfi,3, cfcolor[k]);
						}

						obj->DeleteFace(j);
						divided = true;
						break;
					}
					if(nvi[k]>=0 && nvi[(k+1)%4]>=0)
					{
						MQCoordinate t0,t1;
						DWORD col0,col1;
						if(cfvi[k] < cfvi[(k+1)%4]){
							t0 = cfcoord[k]*(1-t[k]) + cfcoord[(k+1)%4]*t[k];
							col0 = (MQRGBAColor(cfcolor[k])*(1-t[k]) + MQRGBAColor(cfcolor[(k+1)%4])*t[k]).GetRGBA();
						}else{
							t0 = cfcoord[k]*t[k] + cfcoord[(k+1)%4]*(1-t[k]);
							col0 = (MQRGBAColor(cfcolor[k])*t[k] + MQRGBAColor(cfcolor[(k+1)%4])*(1-t[k])).GetRGBA();
						}
						if(cfvi[(k+1)%4] < cfvi[(k+2)%4]){
							t1 = cfcoord[(k+1)%4]*(1-t[(k+1)%4]) + cfcoord[(k+2)%4]*t[(k+1)%4];
							col1 = (MQRGBAColor(cfcolor[(k+1)%4])*(1-t[(k+1)%4]) + MQRGBAColor(cfcolor[(k+2)%4])*t[(k+1)%4]).GetRGBA();
						}else{
							t1 = cfcoord[(k+1)%4]*t[(k+1)%4] + cfcoord[(k+2)%4]*(1-t[(k+1)%4]);
							col1 = (MQRGBAColor(cfcolor[(k+1)%4])*t[(k+1)%4] + MQRGBAColor(cfcolor[(k+2)%4])*(1-t[(k+1)%4])).GetRGBA();
						}

						if (newobj_vi[cfvi[(k+1)%4]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_vi[cfvi[(k+1)%4]];
							nfvert[2] = newobj_nvi[(k+1)%4];
							nfi = newobj->AddFace(3, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%4];
							nfcoord[2] = t1;
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%4]);
							newobj->SetFaceVertexColor(nfi,2, col1);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = cfvi[(k+1)%4];
							nfvert[2] = nvi[(k+1)%4];
							nfi = obj->AddFace(3, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+1)%4];
							nfcoord[2] = t1;
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, cfcolor[(k+1)%4]);
							obj->SetFaceVertexColor(nfi,2, col1);
						}

						if (newobj_vi[cfvi[(k+2)%4]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_nvi[(k+1)%4];
							nfvert[2] = newobj_vi[cfvi[(k+2)%4]];
							nfvert[3] = newobj_vi[cfvi[(k+3)%4]];
							nfi = newobj->AddFace(4, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+2)%4];
							nfcoord[3] = cfcoord[(k+3)%4];
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, col1);
							newobj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%4]);
							newobj->SetFaceVertexColor(nfi,3, cfcolor[(k+3)%4]);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = nvi[(k+1)%4];
							nfvert[2] = cfvi[(k+2)%4];
							nfvert[3] = cfvi[(k+3)%4];
							nfi = obj->AddFace(4, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = t1;
							nfcoord[2] = cfcoord[(k+2)%4];
							nfcoord[3] = cfcoord[(k+3)%4];
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, col1);
							obj->SetFaceVertexColor(nfi,2, cfcolor[(k+2)%4]);
							obj->SetFaceVertexColor(nfi,3, cfcolor[(k+3)%4]);
						}

						if (newobj_vi[cfvi[k]] != -1){
							nfvert[0] = newobj_nvi[k];
							nfvert[1] = newobj_vi[cfvi[(k+3)%4]];
							nfvert[2] = newobj_vi[cfvi[k]];
							nfi = newobj->AddFace(3, nfvert);
							newobj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+3)%4];
							nfcoord[2] = cfcoord[k];
							newobj->SetFaceCoordinateArray(nfi, nfcoord);
							newobj->SetFaceVertexColor(nfi,0, col0);
							newobj->SetFaceVertexColor(nfi,1, cfcolor[(k+3)%4]);
							newobj->SetFaceVertexColor(nfi,2, cfcolor[k]);
						}else{
							nfvert[0] = nvi[k];
							nfvert[1] = cfvi[(k+3)%4];
							nfvert[2] = cfvi[k];
							nfi = obj->AddFace(3, nfvert);
							obj->SetFaceMaterial(nfi, pal);
							nfcoord[0] = t0;
							nfcoord[1] = cfcoord[(k+3)%4];
							nfcoord[2] = cfcoord[k];
							obj->SetFaceCoordinateArray(nfi, nfcoord);
							obj->SetFaceVertexColor(nfi,0, col0);
							obj->SetFaceVertexColor(nfi,1, cfcolor[(k+3)%4]);
							obj->SetFaceVertexColor(nfi,2, cfcolor[k]);
						}

						obj->DeleteFace(j);
						divided = true;
						break;
					}
				}
				break;
			}

			// 分断されない面の移動処理
			if (!divided)
			{
				// 分離側にある面かどうか
				for(k=0; k<cf_count; k++){
					if (newobj_vi[cfvi[k]] == -1)
						break;
					
					nfvert[k] = newobj_vi[cfvi[k]];
				}

				// すべての頂点が分離側にあれば、面を移動
				if (k == cf_count){
					nfi = newobj->AddFace(cf_count, nfvert);
					newobj->SetFaceMaterial(nfi, pal);
					newobj->SetFaceCoordinateArray(nfi, cfcoord);
					for(k=0; k<cf_count; k++){
						newobj->SetFaceVertexColor(nfi,k, cfcolor[k]);
					}

					obj->DeleteFace(j);
				}
			}
		}

		free(hash);
		delete[] cut_face;
		delete[] visible;
		delete[] newobj_vi;

		// 新規に生成した分、インデックスを+1
		oi++;
	}
}



//---------------------------------------------------------------------------
//  GetPluginClass
//    プラグインのベースクラスを返す
//---------------------------------------------------------------------------
MQBasePlugin *GetPluginClass()
{
	static ScissorPlugin plugin;
	return &plugin;
}



//---------------------------------------------------------------------------
//  DllMain
//---------------------------------------------------------------------------
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    return TRUE;
}

