﻿//---------------------------------------------------------------------------
//
//   WTLTestPlugin
//
//          Copyright(C) 2006-2012, O.Mizno
//
//    　常駐プラグインのサンプルです。
//    　MQOファイル読み書き時のXMLファイル(.mqx)の操作、および描画オブ
//    ジェクトを常駐プラグインからどのように使うかを示しています。
//    　ビルドにはWTL(Windows Template Library)が別途必要です。
//
//---------------------------------------------------------------------------
#include "stdafx.h"
#include "resource.h"
#include "maindlg.h"
#include "WTLTestPlugin.h"

static WTLTestPlugin plugin;


// コンストラクタ
WTLTestPlugin::WTLTestPlugin()
{
	dlgMain = NULL;
	m_bActivate = false;
	m_bShowLocal = false;
}

// コンストラクタ
WTLTestPlugin::~WTLTestPlugin()
{
	if (dlgMain != NULL){
		// Exit()で破棄されるはずだが一応念のためここでも
		dlgMain->DestroyWindow();
		delete dlgMain;
		dlgMain = NULL;
	}
}

//---------------------------------------------------------------------------
//  GetPlugInID
//    プラグインIDを返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
void WTLTestPlugin::GetPlugInID(DWORD *Product, DWORD *ID)
{
	// プロダクト名(制作者名)とIDを、全部で64bitの値として返す
	// 値は他と重複しないようなランダムなもので良い
	*Product = 0x56A31D20;
	*ID      = 0x6A4102BF;
}

//---------------------------------------------------------------------------
//  GetPlugInName
//    プラグイン名を返す。
//    この関数は[プラグインについて]表示時に呼び出される。
//---------------------------------------------------------------------------
const char *WTLTestPlugin::GetPlugInName(void)
{
	return "Test station with WTL       Copyright(C) 2006-2012, O.Mizno";
}

//---------------------------------------------------------------------------
//  EnumString
//    ポップアップメニューに表示される文字列を返す。
//    この関数は起動時に呼び出される。
//---------------------------------------------------------------------------
const char *WTLTestPlugin::EnumString(void)
{
	return "WTL station";
}


//---------------------------------------------------------------------------
//  Initialize
//    アプリケーションの初期化
//---------------------------------------------------------------------------
BOOL WTLTestPlugin::Initialize()
{
	if (dlgMain == NULL){
		// アプリケーションにメインウインドウのハンドルを割り当て
		dlgMain = new CMainDlg(this);
		if(dlgMain->Create(MQ_GetWindowHandle()) == NULL)
		{
			delete dlgMain;
			dlgMain = NULL;
			ATLTRACE(_T("Main dialog creation failed!\n"));
			return 0;
		}

		// ウインドウを常に前面に
		dlgMain->SetWindowPos(HWND_TOPMOST, 0,0,0,0,SWP_NOMOVE|SWP_NOSIZE|SWP_NOACTIVATE);
	}
	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//    アプリケーションの終了
//---------------------------------------------------------------------------
void WTLTestPlugin::Exit()
{
	if (dlgMain != NULL){
		dlgMain->DestroyWindow();
		delete dlgMain;
		dlgMain = NULL;
	}
}

//---------------------------------------------------------------------------
//  Activate
//    表示・非表示切り替え要求
//---------------------------------------------------------------------------
BOOL WTLTestPlugin::Activate(MQDocument doc, BOOL flag)
{
	if (dlgMain == NULL)
		return FALSE;

	m_bActivate = flag ? true : false;
	dlgMain->ShowWindow(m_bActivate ? SW_SHOW : SW_HIDE);
	if (m_bActivate){
		//dlgMain->UpdateList(doc);
	}
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//    表示・非表示状態の返答
//---------------------------------------------------------------------------
BOOL WTLTestPlugin::IsActivated(MQDocument doc)
{
	if (dlgMain == NULL)
		return FALSE;

	return m_bActivate;
}

//---------------------------------------------------------------------------
//  OnNewDocument
//    ドキュメント初期化時
//---------------------------------------------------------------------------
void WTLTestPlugin::OnNewDocument(MQDocument doc, const char *filename, NEW_DOCUMENT_PARAM& param)
{
	// XML要素が保存されていない、または新規ドキュメントの場合は何もしない
	if (param.elem == NULL)
		return;

	// 保存された設定を読み込むサンプル
	MQXmlElement child;
	
	child = param.elem->FirstChildElement("ShowLocal");
	if (child != NULL){
		m_bShowLocal = atoi(child->GetText().c_str()) ? TRUE : FALSE;
		dlgMain->m_CheckShowLocal.SetCheck(m_bShowLocal ? 1 : 0);
	}
}

//---------------------------------------------------------------------------
//  OnSaveDocument
//    ドキュメント保存時
//---------------------------------------------------------------------------
void WTLTestPlugin::OnSaveDocument(MQDocument doc, const char *filename, SAVE_DOCUMENT_PARAM& param)
{
	// 設定を保存するためのXML要素を埋め込むサンプル
	MQXmlElement child;
	
	child = param.elem->AddChildElement("ShowLocal");
	child->SetText(m_bShowLocal ? "1" : "0");
}

//---------------------------------------------------------------------------
//  OnDraw
//    描画時の処理
//---------------------------------------------------------------------------
void WTLTestPlugin::OnDraw(MQDocument doc, MQScene scene, int width, int height)
{
	if (!m_bShowLocal)
		return;

	// 注視点を基準に10ピクセル分の長さを算出
	MQPoint sp = scene->Convert3DToScreen(scene->GetLookAtPosition());
	MQPoint p = scene->ConvertScreenTo3D(MQPoint(sp.x+10.0f,sp.y+10.0f,sp.z)) - scene->GetLookAtPosition();
	float len = p.abs();

	// オブジェクトごとのローカル座標の軸を描画
	int numObj = doc->GetObjectCount();
	for (int i=0; i<numObj; i++)
	{
		MQObject obj = doc->GetObject(i);
		if (obj == NULL)
			continue;
		
		MQMatrix mtx;
		doc->GetGlobalMatrix(obj, mtx);

		// 描画オブジェクトを生成
		MQObject axisX = this->CreateDrawingObject(doc, DRAW_OBJECT_LINE);
		MQObject axisY = this->CreateDrawingObject(doc, DRAW_OBJECT_LINE);
		MQObject axisZ = this->CreateDrawingObject(doc, DRAW_OBJECT_LINE);
		axisX->SetColor(MQColor(1,0,0));
		axisY->SetColor(MQColor(0,1,0));
		axisZ->SetColor(MQColor(0,0,1));
		axisX->SetColorValid(TRUE);
		axisY->SetColorValid(TRUE);
		axisZ->SetColorValid(TRUE);

		// 各描画オブジェクトにラインを追加
		int vert[2];
		vert[0] = axisX->AddVertex(MQPoint(0,0,0) * mtx);
		vert[1] = axisX->AddVertex(MQPoint(len,0,0) * mtx);
		axisX->AddFace(2, vert);
		vert[0] = axisY->AddVertex(MQPoint(0,0,0) * mtx);
		vert[1] = axisY->AddVertex(MQPoint(0,len,0) * mtx);
		axisY->AddFace(2, vert);
		vert[0] = axisZ->AddVertex(MQPoint(0,0,0) * mtx);
		vert[1] = axisZ->AddVertex(MQPoint(0,0,len) * mtx);
		axisZ->AddFace(2, vert);
	}
}

//---------------------------------------------------------------------------
//  OnObjectModified
//    オブジェクトの編集時
//---------------------------------------------------------------------------
void WTLTestPlugin::OnObjectModified(MQDocument doc)
{
	//m_pForm->UpdateList(doc);
}

//---------------------------------------------------------------------------
//  OnUpdateObjectList
//    カレントオブジェクトの変更時
//---------------------------------------------------------------------------
void WTLTestPlugin::OnUpdateObjectList(MQDocument doc)
{
	//m_pForm->UpdateList(doc);
}

//---------------------------------------------------------------------------
//  ExecuteCallback
//    コールバックに対する実装部
//---------------------------------------------------------------------------
bool WTLTestPlugin::ExecuteCallback(MQDocument doc, void *option)
{
	CallbackInfo *info = (CallbackInfo*)option;
	return ((*this).*info->proc)(doc);
}

// コールバックの呼び出し
void WTLTestPlugin::Execute(ExecuteCallbackProc proc)
{
	CallbackInfo info;
	info.proc = proc;
	BeginCallback(&info);
}


// 面を消す
bool WTLTestPlugin::DeleteFace(MQDocument doc)
{
	// カレントオブジェクトを取得
	MQSelectFace sf;
	sf.object = doc->GetCurrentObjectIndex();
	MQObject obj = doc->GetObject(sf.object);
	if (obj == NULL)
		return false;

	// 選択されている面を消す
	int numFace = obj->GetFaceCount();
	for (sf.face=0; sf.face<numFace; sf.face++){
		if (doc->IsSelectFace(sf)){
			obj->DeleteFace(sf.face);
		}
	}

	return true;
}


// プラグインのベースクラスを返す
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

